/**
 * Copyright (C) 2012-2017  Dušan Vejnovič  <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.dussan.vaadin.dtable.client.components;

import static com.google.gwt.query.client.GQuery.$;
import static com.google.gwt.query.client.plugins.Effects.Effects;

import org.dussan.vaadin.dtable.client.Translation;
import org.dussan.vaadin.dtable.client.VDTable;
import org.dussan.vaadin.dtable.client.components.base.TableComponent;
import org.dussan.vaadin.dtable.client.components.events.QuickSearchEvent;

import com.google.gwt.dom.client.Element;
import com.google.gwt.query.client.Function;
import com.google.gwt.query.client.GQuery;
import com.google.gwt.query.client.css.CSS;
import com.google.gwt.query.client.css.Length;
import com.google.gwt.user.client.DOM;
import com.google.gwt.user.client.Event;

public class QuickSearch
    extends TableComponent
{

    private boolean hidden = true;

    private String mainId = "div#v-dtable";

    private String quickSearchId = "#dt-quickSearch";

    private Boolean messageSend = false;

    private String focusValue = new String();

    private String eventValue = new String();

    private GQuery focusEventHandler = null;

    private GQuery blurEventHandler = null;

    private GQuery keyDownEventHandler = null;

    private GQuery keyUpEventHandler = null;

    public QuickSearch( VDTable dTable )
    {
        super( dTable );
    }

    private QuickSearch getThis()
    {
        return this;
    }

    private boolean isHidden()
    {
        return hidden;
    }

    private void setHidden( boolean hidden )
    {
        this.hidden = hidden;
    }

    private void activateFocusEventHandler( boolean activate )
    {
        if ( activate )
        {
            activateFocusEventHandler( false );
            focusEventHandler = $( quickSearchId ).focus( new Function()
            {
                public boolean f( Event event )
                {
                    return QuickSearchEvent.focus( getThis(), event );
                }
            } );
        }
        else if ( focusEventHandler != null )
        {
            focusEventHandler.off();
            focusEventHandler = null;
        }
    }

    private void activateBlurEventHandler( boolean activate )
    {
        if ( activate )
        {
            activateBlurEventHandler( false );
            blurEventHandler = $( quickSearchId ).blur( new Function()
            {
                public boolean f( Event event )
                {
                    return QuickSearchEvent.blur( getThis(), event );
                }
            } );
        }
        else if ( blurEventHandler != null )
        {
            blurEventHandler.off();
            blurEventHandler = null;
        }
    }

    private void activateKeyDownEventHandler( boolean activate )
    {
        if ( activate )
        {
            activateKeyDownEventHandler( false );
            keyDownEventHandler = $( quickSearchId ).keydown( new Function()
            {
                public boolean f( Event event )
                {
                    return QuickSearchEvent.keyDown( getThis(), event );
                }
            } );
        }
        else if ( keyDownEventHandler != null )
        {
            keyDownEventHandler.off();
            keyDownEventHandler = null;
        }
    }

    private void activateKeyUpEventHandler( boolean activate )
    {
        if ( activate )
        {
            activateKeyUpEventHandler( false );
            keyUpEventHandler = $( quickSearchId ).keyup( new Function()
            {
                public boolean f( Event event )
                {
                    return QuickSearchEvent.keyUp( getThis(), event );
                }
            } );
        }
        else if ( keyUpEventHandler != null )
        {
            keyUpEventHandler.off();
            keyUpEventHandler = null;
        }
    }

    private void show( boolean animate )
    {
        if ( !isVisible() && !isHidden() )
        {
            Element quickSearchDiv = DOM.createDiv();
            quickSearchDiv.setClassName( "dt-quick-search" );
            $( mainId ).find( "#dt-main-container" )
                       .before( quickSearchDiv );

            Element quickSearchLabel = DOM.createLabel();
            quickSearchLabel.setInnerText( getDTable().translation()
                                                      .get( Translation.SEARCH ) );
            quickSearchDiv.appendChild( quickSearchLabel );

            Element quickSearchBox = DOM.createInputText();
            quickSearchBox.setId( quickSearchId.substring( 1 ) );
            quickSearchDiv.appendChild( quickSearchBox );

            if ( animate )
            {
                int height = quickSearchDiv.getClientHeight();
                $( quickSearchDiv ).css( CSS.HEIGHT.with( Length.px( 0 ) ) );
                String animation = "height:" + height + "px;";

                $( quickSearchDiv ).as( Effects )
                                   .animate( animation, getDTable().config()
                                                                   .getQuickSearchAnimationDuration(),
                                             new Function()
                                             {
                                                 public void f( Element e )
                                                 {
                                                     e.getStyle()
                                                      .clearHeight();
                                                 }
                                             } );
            }

            $( quickSearchId ).parent()
                              .parent()
                              .css( "-moz-border-radius", "" )
                              .css( "-webkit-border-radius", "" )
                              .css( "-khtml-border-radius", "" )
                              .css( "border-radius", "" );
            setEventsHandlerEnabled();
        }
    }

    private boolean isVisible()
    {
        return 0 < $( quickSearchId ).size();
    }

    public void emptyQuickSearch()
    {
        focusValue = new String();
        eventValue = new String();
        $( quickSearchId ).val( "" );
    }

    public Boolean isMessageSend()
    {
        return messageSend;
    }

    public void setMessageSend( Boolean messageSend )
    {
        this.messageSend = messageSend;
    }

    public String getFocusValue()
    {
        return focusValue;
    }

    public void setFocusValue( String focusValue )
    {
        this.focusValue = focusValue;
    }

    public String getEventValue()
    {
        return eventValue;
    }

    public void setEventValue( String eventValue )
    {
        this.eventValue = eventValue;
    }

    @Override
    public void setIds()
    {
        if ( !mainId.endsWith( Long.toString( getDTable().getId() ) ) )
        {
            mainId += Long.toString( getDTable().getId() );
            quickSearchId += Long.toString( getDTable().getId() );

            setHidden( !getDTable().config()
                                   .isQuickSearchVisibleAtStartupEnabled() );
        }
    }

    @Override
    public void load()
    {
        setIds();
        if ( getDTable().config()
                        .isQuickSearchEnabled()
            && getDTable().config()
                          .isQuickSearchVisibleAtStartupEnabled() )
        {
            show( false );
        }
        else
        {
            unload();
        }
    }

    @Override
    public void unload()
    {
        if ( isVisible() )
        {
            String animate = "height:0px;";
            $( quickSearchId ).parent()
                              .as( Effects )
                              .animate( animate, getDTable().config()
                                                            .getQuickSearchAnimationDuration(),
                                        new Function()
                                        {
                                            public void f( Element e )
                                            {
                                                $( mainId ).children( "#dt-main-container-wrapper" )
                                                           .css( "-moz-border-radius", "0" )
                                                           .css( "-webkit-border-radius", "0" )
                                                           .css( "-khtml-border-radius", "0" )
                                                           .css( "border-radius", "0" );
                                                e.removeFromParent();

                                                // send message
                                                String value = eventValue.trim();
                                                if ( !value.isEmpty() && 0 < value.length() )
                                                {
                                                    getDTable().sendMessage( "quickSearch", "", true );
                                                }
                                            }
                                        } );
        }
        setEventsHandlerDisabled();
    }

    public void toggle()
    {
        if ( isVisible() )
        {
            setHidden( true );
            unload();
        }
        else
        {
            setHidden( false );
            show( true );
        }
    }

    @Override
    public void setEventsHandlerEnabled()
    {
        if ( getDTable().config()
                        .isQuickSearchEnabled() )
        {
            activateFocusEventHandler( true );
            activateBlurEventHandler( true );
            activateKeyDownEventHandler( true );
            activateKeyUpEventHandler( true );
        }
    }

    @Override
    public void setEventsHandlerDisabled()
    {
        activateFocusEventHandler( false );
        activateBlurEventHandler( false );
        activateKeyDownEventHandler( false );
        activateKeyUpEventHandler( false );
    }

}
