/**
 * Copyright (C) 2012-2017  Dušan Vejnovič  <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.dussan.vaadin.dtable.client.components.table;

import static com.google.gwt.query.client.GQuery.$;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.dussan.vaadin.dtable.client.VDTable;
import org.dussan.vaadin.dtable.client.components.Tooltip;
import org.dussan.vaadin.dtable.client.components.base.TableComponent;
import org.dussan.vaadin.dtable.client.components.events.TableEvent;

import com.google.gwt.dom.client.Element;
import com.google.gwt.query.client.Function;
import com.google.gwt.query.client.GQuery;
import com.google.gwt.query.client.css.CSS;
import com.google.gwt.query.client.css.Length;
import com.google.gwt.query.client.css.RGBColor;
import com.google.gwt.query.client.css.TakesCssValue.CssSetter;
import com.google.gwt.user.client.Event;
import com.vaadin.client.BrowserInfo;

public class TableHeader
    extends TableComponent
{

    private boolean show = false;

    private String frozenHeaderId = "table#dt-frozen-header-table";

    private String headerId = "table#dt-header-table";

    private List<Integer> columns = new ArrayList<>();

    private Map<Integer, Integer> columnsAlignment = new HashMap<>();

    private Map<Integer, Integer> columnsWidth = new HashMap<>();

    private Map<Integer, String> headers = new HashMap<>();

    private Map<Integer, String> tooltips = new HashMap<>();

    private GQuery cellClickEventHandler = null;

    private GQuery cellContextMenuEventHandler = null;

    private GQuery cellMouseOverEventHandler = null;

    private GQuery frozenCellMouseOverEventHandler = null;

    private GQuery frozenCellClickEventHandler = null;

    private GQuery frozenCellContextMenuEventHandler = null;

    private Tooltip tooltip = null;

    public TableHeader( VDTable dTable )
    {
        super( dTable );
        tooltip = new Tooltip();
    }

    public Tooltip tooltip()
    {
        return tooltip;
    }

    private void setColumns( Object[] columns )
    {
        this.columns.clear();
        for ( Object column : columns )
        {
            this.columns.add( Integer.parseInt( column.toString() ) );
        }
    }

    private void setColumnsWidth( Object[] widths )
    {
        columnsWidth.clear();
        for ( int column = 0; column < widths.length; column++ )
        {
            columnsWidth.put( columns.get( column ), Integer.parseInt( widths[column].toString() ) );
        }
    }

    private void setColumnsWidth( Map<Integer, Integer> columnsWidth )
    {
        if ( columnsWidth != null && columnsWidth.size() > 0 )
        {
            this.columnsWidth = new HashMap<>( columnsWidth );
        }
    }

    public String getHeader( int column )
    {
        return headers.get( column );
    }

    private void setHeaders( Object[] headers )
    {
        if ( headers != null && headers.length > 0 )
        {
            this.headers.clear();
            for ( int column = 0; column < headers.length; column++ )
            {
                this.headers.put( columns.get( column ), headers[column].toString() );
            }
        }
    }

    public String getTooltip( int column )
    {
        return tooltips.get( column );
    }

    private void setTooltips( Object[] tooltips )
    {
        if ( tooltips != null && tooltips.length > 0 )
        {
            this.tooltips.clear();
            for ( int column = 0; column < tooltips.length; column++ )
            {
                this.tooltips.put( columns.get( column ), tooltips[column].toString() );
            }
        }
    }

    private void setColumnsAlignment( Object[] columnsAlignment )
    {
        if ( columnsAlignment != null && columnsAlignment.length > 0 )
        {
            this.columnsAlignment.clear();
            for ( int column = 0; column < columnsAlignment.length; column++ )
            {
                this.columnsAlignment.put( columns.get( column ),
                                           Integer.parseInt( columnsAlignment[column].toString() ) );
            }
        }
    }

    private void emptyHeader()
    {
        $( headerId ).empty()
                     .append( $( "<tr/>" ).id( getDTable().getId() + "h" ) );
        if ( getDTable().config()
                        .isFrozenTableEnabled() )
        {
            int width = getDTable().isFrozenTableHidden() ? 0 : getDTable().getFrozenTableWidth();
            $( frozenHeaderId ).empty()
                               .append( $( "<tr/>" ).id( getDTable().getId() + "fh" ) );
            $( frozenHeaderId ).parent()
                               .parent()
                               .css( CSS.WIDTH.with( Length.px( width ) ) );
        }
        else
        {
            $( frozenHeaderId ).remove();
        }
    }

    private int getColumnWidth( int column )
    {
        int columnWidth = columnsWidth.get( column ) - getDTable().getBrowserOffset();
        if ( BrowserInfo.get()
                        .isChrome() )
        {
            columnWidth -= 1;
        }
        return columnWidth;
    }

    public int getColumnHeight()
    {
        return getDTable().config()
                          .getHeaderRowHeight()
            - getDTable().config()
                         .getHeaderCellMarginTop()
            - getDTable().config()
                         .getHeaderCellMarginBottom()
            - 1;
    }

    private void fillHeader()
    {
        int controledWidth = getDTable().getLeft() - $( headerId ).parent()
                                                                  .scrollLeft();
        int tableWidth = BrowserInfo.get()
                                    .getScreenWidth();
        String clazz = getDTable().config()
                                  .isWordWrapInHeaderCellsEnabled() ? "normal" : "nowrap";
        String id = getDTable().getId() + "h:0:";

        CssSetter height = CSS.HEIGHT.with( Length.px( getColumnHeight() ) );
        CssSetter lineHeight = CSS.LINE_HEIGHT.with( Length.px( getDTable().config()
                                                                           .getHeaderLineHeight() ) );
        CssSetter margins = CSS.MARGIN.with( Length.px( getDTable().config()
                                                                   .getHeaderCellMarginTop() ),
                                             Length.px( getDTable().config()
                                                                   .getHeaderCellMarginRight() ),
                                             Length.px( getDTable().config()
                                                                   .getHeaderCellMarginBottom() ),
                                             Length.px( getDTable().config()
                                                                   .getHeaderCellMarginLeft() ) );

        for ( int column : columns )
        {
            if ( !( tableWidth < controledWidth ) || getDTable().isFrozen( column ) )
            {
                CssSetter align = CSS.TEXT_ALIGN.with( VDTable.COLUMN_ALIGN[columnsAlignment.get( column )] );
                CssSetter width = CSS.WIDTH.with( Length.px( getColumnWidth( column ) ) );
                $( "<div/>" ).id( id + column + ":1" )
                             .css( height )
                             .addClass( clazz )
                             .css( margins )
                             .append( $( "<span/>" ).id( id + column + ":2" )
                                                    .css( lineHeight )
                                                    .text( headers.get( column ) ) )
                             .appendTo( $( "<td/>" ).id( id + column )
                                                    .css( align )
                                                    .css( width )
                                                    .appendTo( $( headerId ).children( "tr" ) ) );

                // only display columns which are in screen area
                if ( !getDTable().isFrozen( column ) && !( getDTable().config()
                                                                      .isAllHeaderColumnsEnabled()
                    || getDTable().config()
                                  .isAllColumnsEnabled() ) )
                {
                    controledWidth += columnsWidth.get( column );
                    if ( tableWidth < controledWidth )
                    {
                        getDTable().setAllColumnsVisible( false );
                    }
                }
            }
        }
    }

    private void setFrozenHeader()
    {
        if ( getDTable().config()
                        .isFrozenTableEnabled() )
        {
            String id = "td [id='" + getDTable().getId() + "h:0:_']";
            if ( columns != null )
            {
                for ( int column : columns )
                {
                    if ( getDTable().isFrozen( column ) )
                    {
                        GQuery cell = $( headerId ).children( "tr" )
                                                   .children( id.replace( "_", Integer.toString( column ) ) );
                        $( frozenHeaderId ).children( "tr" )
                                           .append( cell.clone() );
                        cell.remove();
                    }
                }
            }
            removeRightBorderFromLastColumn( frozenHeaderId );
        }
    }

    private void removeExpandedCells()
    {
        if ( getDTable().config()
                        .isFrozenTableEnabled() )
        {
            $( frozenHeaderId ).children( "tr" )
                               .eq( 0 )
                               .find( "div#clone" )
                               .parent()
                               .children( "div.normal", "div.nowrap" )
                               .css( CSS.DISPLAY, "" )
                               .parent()
                               .children( "div#clone" )
                               .remove();
        }

        $( headerId ).children( "tr" )
                     .eq( 0 )
                     .find( "div#clone" )
                     .parent()
                     .children( "div.normal", "div.nowrap" )
                     .css( CSS.DISPLAY, "" )
                     .parent()
                     .children( "div#clone" )
                     .remove();
    }

    private void removeVerticalLines()
    {
        if ( !getDTable().config()
                         .isHeaderVerticalLinesEnabled() )
        {
            $( headerId ).children( "tr" )
                         .children( "td" )
                         .css( CSS.BORDER_LEFT_COLOR.with( RGBColor.TRANSPARENT ) );
            $( headerId ).children( "tr" )
                         .children( "td" )
                         .css( CSS.BORDER_RIGHT_COLOR.with( RGBColor.TRANSPARENT ) );
        }
    }

    private void removeRightBorderFromLastColumn( String header )
    {
        $( header ).children( "tr" )
                   .children( "td" )
                   .last()
                   .css( CSS.BORDER_RIGHT_COLOR.with( RGBColor.TRANSPARENT ) );
    }

    private void activateFrozenCellMouseOverEventHandler( boolean enable )
    {
        if ( enable )
        {
            activateFrozenCellMouseOverEventHandler( false );

            GQuery header = $( "table#dt-frozen-header-table" + getDTable().getId() );
            frozenCellMouseOverEventHandler = header.on( "mouseover", "td", new Function()
            {
                public boolean f( Event e )
                {
                    if ( getDTable().config()
                                    .isExpandedFrozenHeaderCellEnabled() )
                    {
                        TableEvent.cellMouseOver( getDTable(), $( e ), null );
                    }

                    String id = $( e ).id();
                    if ( e.getRelatedEventTarget() != null && id != null && !id.isEmpty() )
                    {
                        String[] ids = id.split( ":" );
                        if ( getDTable().config()
                                        .isTooltipEnabled()
                            && !getDTable().contextMenu()
                                           .isShowing()
                            && 3 <= ids.length && "0".equals( ids[1] ) )
                        {
                            int column = Integer.parseInt( id.split( ":" )[2] );
                            tooltip.setTo( e, getTooltip( column ), //
                                           getDTable().config()
                                                      .isWordWrapInTooltipEnabled(),
                                           getDTable().config()
                                                      .getTooltipMinimumWidth(),
                                           getDTable().config()
                                                      .getTooltipMaximumWidth(),
                                           getDTable().config()
                                                      .getTooltipPosition() );
                            tooltip.show();
                        }
                        else
                        {
                            tooltip.hide();
                        }
                    }
                    else
                    {
                        tooltip.hide();
                    }

                    return true;
                }
            } )
                                                    .on( "mouseout", "td", new Function()
                                                    {
                                                        public boolean f( Event e )
                                                        {
                                                            boolean out = TableEvent.headerCellMouseOut( e );
                                                            if ( getDTable().config()
                                                                            .isTooltipEnabled()
                                                                && out )
                                                            {
                                                                tooltip.hide();
                                                            }
                                                            return out;
                                                        }
                                                    } );
        }
        else if ( frozenCellMouseOverEventHandler != null )
        {
            frozenCellMouseOverEventHandler.off( "mouseover", "td" )
                                           .off( "mouseout", "td" );
            frozenCellMouseOverEventHandler = null;
        }
    }

    private void activateFrozenCellClickEventHandler( boolean enable )
    {
        if ( enable )
        {
            activateFrozenCellClickEventHandler( false );

            GQuery header = $( "table#dt-frozen-header-table" + getDTable().getId() );
            frozenCellClickEventHandler = header.on( "click", "td", new Function()
            {
                public void f( Element e )
                {
                    if ( getDTable().config()
                                    .isOrderedFrozenTableEnabled()
                        && getDTable().config()
                                      .isTooltipEnabled() )
                    {
                        tooltip.hide();
                    }
                    TableEvent.headerClick( getDTable(), $( e ) );
                }
            } );
        }
        else if ( frozenCellClickEventHandler != null )
        {
            frozenCellClickEventHandler.off( "click", "td" );
            frozenCellClickEventHandler = null;
        }
    }

    private void activateFrozenCellContextMenuEventHandler( boolean enable )
    {
        if ( enable )
        {
            activateFrozenCellContextMenuEventHandler( false );

            GQuery header = $( "table#dt-frozen-header-table" + getDTable().getId() );
            frozenCellContextMenuEventHandler = header.on( "contextmenu", "td", new Function()
            {
                public boolean f( Event e )
                {
                    if ( getDTable().config()
                                    .isTooltipEnabled() )
                    {
                        tooltip.hide();
                    }

                    return TableEvent.contextMenu( getDTable(), e );
                }
            } );
        }
        else if ( frozenCellContextMenuEventHandler != null )
        {
            frozenCellContextMenuEventHandler.off( "contextmenu", "td" );
            frozenCellContextMenuEventHandler = null;
        }
    }

    private void activateCellMouseOverEventHandler( boolean enable )
    {
        if ( enable )
        {
            activateCellMouseOverEventHandler( false );

            GQuery header = $( "table#dt-header-table" + getDTable().getId() );
            cellMouseOverEventHandler = header.on( "mouseover", "td", new Function()
            {
                public boolean f( Event e )
                {
                    if ( getDTable().config()
                                    .isExpandedHeaderCellEnabled() )
                    {
                        TableEvent.cellMouseOver( getDTable(), $( e ), null );
                    }

                    String id = $( e ).id();
                    if ( e.getRelatedEventTarget() != null && id != null && !id.isEmpty() )
                    {
                        String[] ids = id.split( ":" );
                        if ( getDTable().config()
                                        .isTooltipEnabled()
                            && !getDTable().contextMenu()
                                           .isShowing()
                            && 3 <= ids.length && "0".equals( ids[1] ) )
                        {
                            int column = Integer.parseInt( ids[2] );
                            tooltip.setTo( e, getTooltip( column ), //
                                           getDTable().config()
                                                      .isWordWrapInTooltipEnabled(),
                                           getDTable().config()
                                                      .getTooltipMinimumWidth(),
                                           getDTable().config()
                                                      .getTooltipMaximumWidth(),
                                           getDTable().config()
                                                      .getTooltipPosition() );
                            tooltip.show();
                        }
                        else
                        {
                            tooltip.hide();
                        }
                    }
                    else
                    {
                        tooltip.hide();
                    }
                    return true;
                }
            } )
                                              .on( "mouseout", "td", new Function()
                                              {
                                                  public boolean f( Event e )
                                                  {
                                                      boolean out = TableEvent.headerCellMouseOut( e );
                                                      if ( getDTable().config()
                                                                      .isTooltipEnabled()
                                                          && out )
                                                      {
                                                          tooltip.hide();
                                                      }
                                                      return out;
                                                  }
                                              } );
        }
        else if ( cellMouseOverEventHandler != null )
        {
            cellMouseOverEventHandler.off( "mouseover", "td" )
                                     .off( "mouseout", "td" );
            cellMouseOverEventHandler = null;
        }
    }

    private void activateCellClickEventHandler( boolean enable )
    {
        if ( enable )
        {
            activateCellClickEventHandler( false );

            GQuery header = $( "table#dt-header-table" + getDTable().getId() );
            cellClickEventHandler = header.on( "click", "td", new Function()
            {
                public void f( Element e )
                {
                    if ( getDTable().config()
                                    .isOrderedTableEnabled()
                        && getDTable().config()
                                      .isTooltipEnabled() )
                    {
                        tooltip.hide();
                    }
                    TableEvent.headerClick( getDTable(), $( e ) );
                }
            } );
        }
        else if ( cellClickEventHandler != null )
        {
            cellClickEventHandler.off( "click", "td" );
            cellClickEventHandler = null;
        }
    }

    private void activateCellContextMenuEventHandler( boolean enable )
    {
        if ( enable )
        {
            activateCellContextMenuEventHandler( false );

            GQuery header = $( "table#dt-header-table" + getDTable().getId() );
            cellContextMenuEventHandler = header.on( "contextmenu", "td", new Function()
            {
                public boolean f( Event e )
                {
                    if ( getDTable().config()
                                    .isTooltipEnabled() )
                    {
                        tooltip.hide();
                    }
                    return TableEvent.contextMenu( getDTable(), e );
                }
            } );
        }
        else if ( cellContextMenuEventHandler != null )
        {
            cellContextMenuEventHandler.off( "contextmenu", "td" );
            cellContextMenuEventHandler = null;
        }
    }

    @Override
    public void setIds()
    {
        if ( !headerId.endsWith( Long.toString( getDTable().getId() ) ) )
        {
            frozenHeaderId += Long.toString( getDTable().getId() );
            headerId += Long.toString( getDTable().getId() );
        }
    }

    public void processData( Map<String, String> data )
    {
        if ( data.containsKey( "columnsHeader" ) && data.containsKey( "visibleHeaderColumns" )
            && data.containsKey( "headerColumnsAlignment" ) )
        {
            show = true;
            setColumns( VDTable.toStringArray( data.get( "visibleHeaderColumns" ) ) );
            setHeaders( VDTable.toStringArray( data.get( "columnsHeader" ) ) );
            setColumnsAlignment( VDTable.toStringArray( data.get( "headerColumnsAlignment" ) ) );

            if ( getDTable().config()
                            .isTooltipEnabled()
                && data.containsKey( "headersTooltip" ) )
            {
                setTooltips( VDTable.toStringArray( data.get( "headersTooltip" ) ) );
            }

            if ( data.containsKey( "headerColumnsWidth" ) )
            {
                setColumnsWidth( VDTable.toStringArray( data.get( "headerColumnsWidth" ) ) );
            }
            else
            {
                setColumnsWidth( getDTable().getColumnsWidth() );
            }
        }
        else
        {
            show = false;
        }
    }

    public void set100percentWidth( boolean is100percentWidth )
    {
        $( headerId ).css( "width", is100percentWidth ? "100%" : "0" );
    }

    @Override
    public void load()
    {
        setIds();
        setEventsHandlerEnabled();
    }

    @Override
    public void unload()
    {
        setEventsHandlerDisabled();
    }

    public void show()
    {
        if ( show )
        {
            boolean showSearch = getDTable().search()
                                            .isVisible();
            emptyHeader();
            fillHeader();

            // finalize
            removeVerticalLines();
            removeRightBorderFromLastColumn( headerId );
            setFrozenHeader();

            getDTable().search()
                       .show( showSearch );
        }
    }

    @Override
    public void setEventsHandlerEnabled()
    {
        activateCellClickEventHandler( getDTable().config()
                                                  .isOrderedTableEnabled() );
        activateCellContextMenuEventHandler( true );
        activateCellMouseOverEventHandler( true );

        if ( getDTable().config()
                        .isFrozenTableEnabled() )
        {
            activateFrozenCellClickEventHandler( getDTable().config()
                                                            .isOrderedFrozenTableEnabled() );
            activateFrozenCellContextMenuEventHandler( true );
            activateFrozenCellMouseOverEventHandler( true );
        }
    }

    @Override
    public void setEventsHandlerDisabled()
    {
        removeExpandedCells();
        activateFrozenCellMouseOverEventHandler( false );
        activateFrozenCellClickEventHandler( false );
        activateFrozenCellContextMenuEventHandler( false );
        activateCellMouseOverEventHandler( false );
        activateCellClickEventHandler( false );
        activateCellContextMenuEventHandler( false );
    }

}
