/**
 * Copyright (C) 2012-2017  Dušan Vejnovič  <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.dussan.vaadin.dtable.config;

import java.io.Serializable;
import java.text.Format;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.commons.lang3.ArrayUtils;
import org.dussan.vaadin.dtable.DTable;
import org.dussan.vaadin.dtable.config.formatting.Formatting;
import org.dussan.vaadin.dtable.config.formatting.FormattingHelper;
import org.dussan.vaadin.dtable.config.freeze.ColumnFreeze;
import org.dussan.vaadin.dtable.config.freeze.ColumnUnfreeze;
import org.dussan.vaadin.dtable.config.hide.ColumnHide;
import org.dussan.vaadin.dtable.config.hide.ColumnUnhide;
import org.dussan.vaadin.dtable.config.tablesection.alignments.ColumnAlign;
import org.dussan.vaadin.dtable.config.tablesection.alignments.RowAlign;
import org.dussan.vaadin.dtable.config.tablesection.dimensions.ColumnWidth;

public class TableConfig
    implements Serializable, ColumnWidth, ColumnFreeze, ColumnUnfreeze, ColumnHide, ColumnUnhide, FormattingHelper
{

    private static final long serialVersionUID = -4562404019696839838L;

    public static final Object DEFAULT_VALUE_FOR_HIDDEN_AND_REMOVED_COLUMNS = 0;

    public static final int DEFAULT_TABLE_WIDTH = 4000;

    public static final int DEFAULT_TABLE_HEIGHT = 300;

    public static final int DEFAULT_TABLE_FROZEN_MAXIMUM_COLUMNS = 10;

    public static final int DEFAULT_TABLE_FROZEN_MINIMAL_WIDTH = 150;

    public static final int DEFAULT_TABLE_MINIMAL_WIDTH = 300;

    public static final int DEFAULT_CONTEXT_MENU_DEPTH_FOR_FILTERED_FROZEN_TABLE = 3;

    public static final int DEFAULT_CONTEXT_MENU_DEPTH_FOR_FILTERED_TABLE = 3;

    public static final int DEFAULT_PAGER_PAGES = 15;

    public static final int DEFAULT_PAGE_SIZE = 15;

    public static final FilterInfoPosition DEFAULT_FILTER_INFO_POSITION = FilterInfoPosition.BOTTOM;

    public static final TableToolsPosition DEFAULT_TABLE_TOOLS_POSITION = TableToolsPosition.TOP_RIGHT;

    public static final int DEFAULT_PAGE_REFRESH = 500;

    public static final int DEFAULT_PAGE_RESIZING_ACTION_DELAY = 500;

    public static final int DEFAULT_SCROLLBAR_ACTION_DELAY = 600;

    public static final int DEFAULT_TABLE_TOOLS_MOUSE_CLICK_ACTION_DELAY = 300;

    public static final int DEFAULT_TABLE_TOOLS_TRAY_MOUSE_OVER_ANIMATION_DURATION = 300;

    public static final int DEFAULT_TABLE_TOOLS_TRAY_ANIMATION_DURATION = 300;

    public static final int DEFAULT_QUICK_SEARCH_ANIMATION_DURATION = 300;

    public static final int DEFAULT_SEARCH_ANIMATION_DURATION = 400;

    public static final int DEFAULT_MINIMAL_ANIMATION_DURATION = 300;

    public static final int DEFAULT_MINIMAL_DELAY = 100;

    public static final ColumnAlign DEFAULT_HEADER_COLUMN_ALIGNMENT = ColumnAlign.CENTER;

    public static final RowAlign DEFAULT_HEADER_ROW_ALIGNMENT = RowAlign.MIDDLE;

    public static final int DEFAULT_HEADER_ROW_HEIGHT = 30;

    public static final int DEFAULT_HEADER_LINE_HEIGHT = 22;

    public static final int DEFAULT_HEADER_TOP_BOTTOM_MARGINS = 3;

    public static final int DEFAULT_HEADER_LEFT_RIGHT_MARGINS = 6;

    public static final ColumnAlign DEFAULT_BODY_COLUMN_ALIGNMENT = ColumnAlign.LEFT;

    public static final RowAlign DEFAULT_BODY_ROW_ALIGNMENT = RowAlign.MIDDLE;

    public static final int DEFAULT_BODY_ROW_HEIGHT = 24;

    public static final int DEFAULT_BODY_LINE_HEIGHT = 20;

    public static final int DEFAULT_BODY_TOP_BOTTOM_MARGINS = 2;

    public static final int DEFAULT_BODY_LEFT_RIGHT_MARGINS = 4;

    public static final float DEFAULT_LOADING_MESSAGE_OPACITY = 0.5f;

    public static final int DEFAULT_LOADING_MESSAGE_SHOW_DELAY = 100;

    public static final int DEFAULT_LOADING_MESSAGE_HIDE_DELAY = 400;

    private boolean checkColumnsWidth = true;

    private boolean prepareDataWithOrdersEnabled = false;

    private boolean prepareDataWithHiddenColumnsEnabled = false;

    private boolean prepareDataWithRemovedColumnsEnabled = false;

    private boolean prepareDataWithOriginalValuesForHiddenAndRemovedColumnsEnabled = false;

    private boolean allColumnsEnabled = false;

    private boolean allRowsEnabled = false;

    private boolean fixedTableHeightEnabled = false;

    private boolean pageableTableEnabled = false;

    private boolean horizontalLinesEnabled = false;

    private boolean highlightActiveRowEnabled = false;

    private boolean oddAndEvenColoringRowsEnabled = false;

    private boolean quickSearchEnabled = false;

    private boolean quickSearchFormattingEnabled = false;

    private boolean quickSearchImmediateEnabled = false;

    private boolean quickSearchVisibleAtStartupEnabled = false;

    private boolean quickSearchHideEnabled = false;

    private boolean quickSearchClearOnNoNewDataEnabled = false;

    private boolean searchEnabled = false;

    private boolean searchFormattingEnabled = false;

    private boolean searchImmediateEnabled = false;

    private boolean searchVisibleAtStartupEnabled = false;

    private boolean searchHideEnabled = false;

    private boolean searchClearOnNoNewDataEnabled = false;

    private boolean noNewDataMessageEnabled = false;

    private boolean loadingMessageEnabled = false;

    private boolean columnNumberInLoadingMessageBoxEnabled = false;

    private boolean orderedFrozenTableEnabled = false;

    private boolean orderedTableEnabled = false;

    private boolean filteredFrozenTableEnabled = false;

    private boolean filteredTableEnabled = false;

    private boolean filterInfoEnabled = false;

    private boolean contextMenuForFilteredFrozenTableEnabled = false;

    private boolean contextMenuForFilteredTableEnabled = false;

    private boolean frozenTableEnabled = false;

    private boolean frozenTableVisibleEnabled = false;

    private boolean hideFrozenTableEnabled = false;

    private boolean emptyFrozenTableEnabled = false;

    private boolean resizeFrozenTableEnabled = false;

    private boolean freezeUnfreezeColumnsEnabled = false;

    private boolean freezeUnfreezeColumnsToolEnabled = false;

    private boolean freezeAnyColumnsEnabled = false;

    private boolean unfreezeAnyColumnsEnabled = false;

    private boolean hideUnhideColumnsEnabled = false;

    private boolean hideUnhideColumnsToolEnabled = false;

    private boolean hideAnyColumnsEnabled = false;

    private boolean unhideAnyColumnsEnabled = false;

    private boolean newDataWhenHideUnhideColumnEnabled = false;

    private boolean tableToolsMenuEnabled = false;

    private boolean tableToolsMenuAlwaysVisibleEnabled = false;

    private boolean tableToolsTrayEnabled = false;

    private int pageResizingActionDelay = DEFAULT_PAGE_RESIZING_ACTION_DELAY;

    private int scrollbarActionDelay = DEFAULT_SCROLLBAR_ACTION_DELAY;

    private int quickSearchAnimationDuration = DEFAULT_QUICK_SEARCH_ANIMATION_DURATION;

    private int searchAnimationDuration = DEFAULT_SEARCH_ANIMATION_DURATION;

    private int allColumns = 0;

    private int columnDefaultWidth = ColumnWidth.DEFAULT_COLUMN_WIDTH;

    private int contextMenuDepthForFilteredFrozenTable = DEFAULT_CONTEXT_MENU_DEPTH_FOR_FILTERED_FROZEN_TABLE;

    private int contextMenuDepthForFilteredTable = DEFAULT_CONTEXT_MENU_DEPTH_FOR_FILTERED_TABLE;

    private int loadingMessageHideDelay = DEFAULT_LOADING_MESSAGE_HIDE_DELAY;

    private int loadingMessageShowDelay = DEFAULT_LOADING_MESSAGE_SHOW_DELAY;

    private int pagerPages = DEFAULT_PAGER_PAGES;

    private int pageSize = DEFAULT_PAGE_SIZE;

    private int tableFrozenMaximumColumns = DEFAULT_TABLE_FROZEN_MAXIMUM_COLUMNS;

    private int tableFrozenMinimalWidth = DEFAULT_TABLE_FROZEN_MINIMAL_WIDTH;

    private int tableHeight = DEFAULT_TABLE_HEIGHT;

    private int tableMinimalWidth = DEFAULT_TABLE_MINIMAL_WIDTH;

    private int tableWidth = DEFAULT_TABLE_WIDTH;

    private int tableToolsMouseClickActionDelay = DEFAULT_TABLE_TOOLS_MOUSE_CLICK_ACTION_DELAY;

    private int tableToolsTrayMouseOverAnimationDuration = DEFAULT_TABLE_TOOLS_TRAY_MOUSE_OVER_ANIMATION_DURATION;

    private int tableToolsTrayAnimationDuration = DEFAULT_TABLE_TOOLS_TRAY_ANIMATION_DURATION;

    private float loadingMessageOpacity = DEFAULT_LOADING_MESSAGE_OPACITY;

    private transient Object defaultValueForHiddenAndRemovedcolumns = DEFAULT_VALUE_FOR_HIDDEN_AND_REMOVED_COLUMNS;

    private List<Formatting> formattingColumns = new ArrayList<>();

    private List<String> columnFreezeEnabled = new ArrayList<>();

    private List<String> columnHideEnabled = new ArrayList<>();

    private List<String> columnUnfreezeEnabled = new ArrayList<>();

    private List<String> columnUnhideEnabled = new ArrayList<>();

    private Map<Integer, ColumnType> columnsType = new HashMap<>();

    private Map<Integer, Integer> columnsWidth = new HashMap<>();

    private BodyConfig bodyConfig = new BodyConfig();

    private FilterInfoPosition filterInfoPosition = DEFAULT_FILTER_INFO_POSITION;

    private HeaderConfig headerConfig = new HeaderConfig();

    private Locale locale = null;

    private TableToolsPosition tableToolsPosition = DEFAULT_TABLE_TOOLS_POSITION;

    public TableConfig()
    {
        getHeader().setDefaultColumnAlignment( DEFAULT_HEADER_COLUMN_ALIGNMENT );
        getHeader().setRowAlignment( DEFAULT_HEADER_ROW_ALIGNMENT );
        getHeader().setRowHeight( DEFAULT_HEADER_ROW_HEIGHT );
        getHeader().setLineHeight( DEFAULT_HEADER_LINE_HEIGHT );
        getHeader().setMargins( DEFAULT_HEADER_TOP_BOTTOM_MARGINS, DEFAULT_HEADER_LEFT_RIGHT_MARGINS );

        getBody().setDefaultColumnAlignment( DEFAULT_BODY_COLUMN_ALIGNMENT );
        getBody().setRowAlignment( DEFAULT_BODY_ROW_ALIGNMENT );
        getBody().setRowHeight( DEFAULT_BODY_ROW_HEIGHT );
        getBody().setLineHeight( DEFAULT_BODY_LINE_HEIGHT );
        getBody().setMargins( DEFAULT_BODY_TOP_BOTTOM_MARGINS, DEFAULT_BODY_LEFT_RIGHT_MARGINS );
    }

    private void checkColumnsWidth()
    {
        if ( checkColumnsWidth && allColumns > 0 )
        {
            // check columns number
            List<Integer> columns = Arrays.asList( columnsWidth.keySet()
                                                               .toArray( new Integer[] {} ) );
            Collections.sort( columns );
            if ( !columns.isEmpty() && !( columns.get( 0 ) >= 0 && columns.get( columns.size() - 1 ) < allColumns ) )
            {
                Map<Integer, Integer> widths = new HashMap<>();
                for ( int column : columns )
                {
                    if ( column >= 0 && column < allColumns )
                    {
                        widths.put( column, columnsWidth.get( column ) );
                    }
                }
                columnsWidth = new HashMap<>( widths );
            }

            // add missing columns width
            if ( columnsWidth.size() < allColumns )
            {
                for ( int column = 0; column < allColumns; column++ )
                {
                    if ( !columnsWidth.containsKey( column ) )
                    {
                        columnsWidth.put( column, columnDefaultWidth );
                    }
                }
            }
            checkColumnsWidth = false;
        }
    }

    public void setAllColumns( int allColumns )
    {
        this.allColumns = Math.max( allColumns, this.allColumns );
        getHeader().setAllColumns( this.allColumns );
        getBody().setAllColumns( this.allColumns );
    }

    public HeaderConfig getHeader()
    {
        return headerConfig;
    }

    public BodyConfig getBody()
    {
        return bodyConfig;
    }

    public Locale getLocale()
    {
        return locale;
    }

    public void setLocale( Locale locale )
    {
        this.locale = locale;
    }

    public int getPageResizingActionDelay()
    {
        return pageResizingActionDelay;
    }

    public void setPageResizingActionDelay( int pageResizingActionDelay )
    {
        this.pageResizingActionDelay =
            DEFAULT_MINIMAL_DELAY <= pageResizingActionDelay ? pageResizingActionDelay : this.pageResizingActionDelay;
    }

    public int getScrollbarActionDelay()
    {
        return scrollbarActionDelay;
    }

    public void setScrollbarActionDelay( int scrollbarActionDelay )
    {
        this.scrollbarActionDelay =
            DEFAULT_MINIMAL_DELAY <= scrollbarActionDelay ? scrollbarActionDelay : this.scrollbarActionDelay;
    }

    public int getTableToolsTrayAnimationDuration()
    {
        return tableToolsTrayAnimationDuration;
    }

    public void setTableToolsTrayAnimationDuration( int tableToolsTrayAnimationDuration )
    {
        this.tableToolsTrayAnimationDuration = DEFAULT_MINIMAL_ANIMATION_DURATION <= tableToolsTrayAnimationDuration
                        ? tableToolsTrayAnimationDuration : DEFAULT_TABLE_TOOLS_TRAY_ANIMATION_DURATION;
    }

    public int getTableToolsMouseClickActionDelay()
    {
        return tableToolsMouseClickActionDelay;
    }

    public void setTableToolsMouseClickActionDelay( int tableToolsMouseClickActionDelay )
    {
        this.tableToolsMouseClickActionDelay = DEFAULT_MINIMAL_DELAY <= tableToolsMouseClickActionDelay
                        ? tableToolsMouseClickActionDelay : DEFAULT_TABLE_TOOLS_MOUSE_CLICK_ACTION_DELAY;
    }

    public int getTableToolsTrayMouseOverAnimationDuration()
    {
        return tableToolsTrayMouseOverAnimationDuration;
    }

    public void setTableToolsTrayMouseOverAnimationDuration( int tableToolsTrayMouseOverAnimationDuration )
    {
        this.tableToolsTrayMouseOverAnimationDuration =
            DEFAULT_MINIMAL_ANIMATION_DURATION <= tableToolsTrayMouseOverAnimationDuration
                            ? tableToolsTrayMouseOverAnimationDuration
                            : DEFAULT_TABLE_TOOLS_TRAY_MOUSE_OVER_ANIMATION_DURATION;
    }

    public int getQuickSearchAnimationDuration()
    {
        return quickSearchAnimationDuration;
    }

    public void setQuickSearchAnimationDuration( int quickSearchAnimationDuration )
    {
        this.quickSearchAnimationDuration = DEFAULT_MINIMAL_ANIMATION_DURATION <= quickSearchAnimationDuration
                        ? quickSearchAnimationDuration : DEFAULT_QUICK_SEARCH_ANIMATION_DURATION;
    }

    public int getSearchAnimationDuration()
    {
        return searchAnimationDuration;
    }

    public void setSearchAnimationDuration( int searchAnimationDuration )
    {
        this.searchAnimationDuration = DEFAULT_MINIMAL_ANIMATION_DURATION <= searchAnimationDuration
                        ? searchAnimationDuration : DEFAULT_SEARCH_ANIMATION_DURATION;
    }

    public int getTableWidth()
    {
        return tableWidth;
    }

    public void setTableWidth( int width )
    {
        tableWidth = width > 0 ? width : tableWidth;
    }

    public int getTableHeight()
    {
        return tableHeight;
    }

    public void setTableHeight( int height )
    {
        // table height minus 2 (2 = borders)
        tableHeight = 0 < height ? ( height - 2 ) : DEFAULT_TABLE_HEIGHT;
    }

    public int getTableFrozenMaximumColumns()
    {
        return tableFrozenMaximumColumns;
    }

    public void setTableFrozenMaximumColumns( int maximumColumns )
    {
        tableFrozenMaximumColumns = 0 < maximumColumns ? maximumColumns : tableFrozenMaximumColumns;
    }

    public int getTableFrozenMinimalWidth()
    {
        return tableFrozenMinimalWidth;
    }

    public void setTableFrozenMinimalWidth( int width )
    {
        tableFrozenMinimalWidth = 0 < width ? width : tableFrozenMinimalWidth;
    }

    public int getTableMinimalWidth()
    {
        return tableMinimalWidth;
    }

    public void setTableMinimalWidth( int width )
    {
        tableMinimalWidth = 0 < width ? width : tableMinimalWidth;
    }

    public int getContextMenuDepthForFilteredFrozenTable()
    {
        return contextMenuDepthForFilteredFrozenTable;
    }

    public void setContextMenuDepthForFilteredFrozenTable( int contextMenuDepthForFilteredFrozenTable )
    {
        this.contextMenuDepthForFilteredFrozenTable =
            Math.max( DEFAULT_CONTEXT_MENU_DEPTH_FOR_FILTERED_FROZEN_TABLE, contextMenuDepthForFilteredFrozenTable );
    }

    public int getContextMenuDepthForFilteredTable()
    {
        return contextMenuDepthForFilteredTable;
    }

    public void setContextMenuDepthForFilteredTable( int contextMenuDepthForFilteredTable )
    {
        this.contextMenuDepthForFilteredTable =
            Math.max( DEFAULT_CONTEXT_MENU_DEPTH_FOR_FILTERED_TABLE, contextMenuDepthForFilteredTable );
    }

    public int getPagerPages()
    {
        return pagerPages;
    }

    public void setPagerPages( int pages )
    {
        pagerPages = 0 < pages ? pages : pagerPages;
    }

    public int getPageSize()
    {
        pageSize = !isPageableTableEnabled() ? ( getTableHeight() / getBody().getRowHeight() + 2 ) : pageSize;
        return pageSize;
    }

    public void setPageSize( int size )
    {
        pageSize = 0 < size ? size : pageSize;
    }

    public String getTableToolsPosition()
    {
        return tableToolsPosition.getPosition();
    }

    public void setTableToolsPosition( TableToolsPosition tableToolsPosition )
    {
        this.tableToolsPosition = tableToolsPosition;
    }

    public String getFilterInfoPosition()
    {
        return filterInfoPosition.getPosition();
    }

    public void setFilterInfoPosition( FilterInfoPosition filterInfoPosition )
    {
        this.filterInfoPosition = filterInfoPosition;
    }

    public Class<?> getColumnTypeClass( int column )
    {
        if ( columnsType.containsKey( column ) )
        {
            return columnsType.get( column )
                              .getType();
        }
        return ColumnType.TEXT.getType();
    }

    public ColumnType getColumnType( int column )
    {
        if ( columnsType.containsKey( column ) )
        {
            return columnsType.get( column );
        }
        return ColumnType.TEXT;
    }

    public void setColumnType( int column, ColumnType columnType )
    {
        columnsType.put( column, columnType );
    }

    public Map<Integer, ColumnType> getColumnsType()
    {
        return columnsType;
    }

    public void setColumnsType( ColumnType columnType, int... columns )
    {
        for ( int column : columns )
        {
            setColumnType( column, columnType );
        }
    }

    @Override
    public int getColumnDefaultWidth()
    {
        return columnDefaultWidth;
    }

    @Override
    public void setColumnDefaultWidth( int width )
    {
        columnDefaultWidth = 0 < width ? width : columnDefaultWidth;
    }

    @Override
    public int getColumnWidth( int column )
    {
        return columnsWidth.containsKey( column ) ? columnsWidth.get( column ) : columnDefaultWidth;
    }

    @Override
    public void setColumnWidth( int column, int width )
    {
        checkColumnsWidth = true;
        columnsWidth.put( column, 0 < width ? width : columnDefaultWidth );
    }

    @Override
    public Map<Integer, Integer> getColumnWidthsMap()
    {
        checkColumnsWidth();
        return columnsWidth;
    }

    @Override
    public int[] getColumnsWidth( int[] columns )
    {
        checkColumnsWidth();
        List<Integer> widths = new ArrayList<>();
        for ( int column : columns )
        {
            widths.add( columnsWidth.get( column ) );
        }
        return ArrayUtils.toPrimitive( widths.toArray( new Integer[] {} ) );
    }

    @Override
    public int[] getColumnsWidth()
    {
        checkColumnsWidth();
        int[] widths = new int[allColumns];
        for ( int column = 0; column < allColumns; column++ )
        {
            widths[column] = columnsWidth.get( column );
        }
        return widths;
    }

    @Override
    public void setColumnsWidth( int... widths )
    {
        checkColumnsWidth = true;
        for ( int column = 0; column < widths.length; column++ )
        {
            setColumnWidth( column, widths[column] );
        }
    }

    @Override
    public void setColumnsSameWidth( int width, int... columns )
    {
        checkColumnsWidth = true;
        for ( int column : columns )
        {
            setColumnWidth( column, width );
        }
    }

    @Override
    public void setColumnsWidth( Map<Integer, Integer> widths )
    {
        checkColumnsWidth = true;
        for ( Entry<Integer, Integer> data : widths.entrySet() )
        {
            Integer column = data.getKey();
            Integer width = data.getValue();
            setColumnWidth( column, width );
        }
    }

    public Object getDefaultValueForHiddenAndRemovedcolumns()
    {
        return defaultValueForHiddenAndRemovedcolumns;
    }

    public void setDefaultValueForHiddenAndRemovedcolumns( Object defaultValueForHiddenAndRemovedcolumns )
    {
        if ( defaultValueForHiddenAndRemovedcolumns != null && !defaultValueForHiddenAndRemovedcolumns.toString()
                                                                                                      .isEmpty()
            && !" ".equals( defaultValueForHiddenAndRemovedcolumns.toString()
                                                                  .trim() ) )
        {
            this.defaultValueForHiddenAndRemovedcolumns = defaultValueForHiddenAndRemovedcolumns;
        }
    }

    public boolean isPrepareDataWithOrdersEnabled()
    {
        return prepareDataWithOrdersEnabled;
    }

    public void setPrepareDataWithOrdersEnabled( boolean prepareDataWithOrdersEnabled )
    {
        this.prepareDataWithOrdersEnabled = prepareDataWithOrdersEnabled;
    }

    public boolean isPrepareDataWithHiddenColumnsEnabled()
    {
        return prepareDataWithHiddenColumnsEnabled;
    }

    public void setPrepareDataWithHiddenColumnsEnabled( boolean prepareDataWithHiddenColumnsEnabled )
    {
        this.prepareDataWithHiddenColumnsEnabled = prepareDataWithHiddenColumnsEnabled;
    }

    public boolean isPrepareDataWithRemovedColumnsEnabled()
    {
        return prepareDataWithRemovedColumnsEnabled;
    }

    public void setPrepareDataWithRemovedColumnsEnabled( boolean prepareDataWithRemovedColumnsEnabled )
    {
        this.prepareDataWithRemovedColumnsEnabled = prepareDataWithRemovedColumnsEnabled;
    }

    public boolean isPrepareDataWithOriginalValuesForHiddenAndRemovedColumnsEnabled()
    {
        return prepareDataWithOriginalValuesForHiddenAndRemovedColumnsEnabled;
    }

    public void setPrepareDataWithOriginalValuesForHiddenAndRemovedColumnsEnabled( boolean prepareDataWithOriginalValuesForHiddenAndRemovedColumnsEnabled )
    {
        this.prepareDataWithOriginalValuesForHiddenAndRemovedColumnsEnabled =
            prepareDataWithOriginalValuesForHiddenAndRemovedColumnsEnabled;
    }

    public boolean isAllColumnsEnabled()
    {
        return allColumnsEnabled;
    }

    public void setAllColumnsEnabled( boolean allColumnsEnabled )
    {
        this.allColumnsEnabled = allColumnsEnabled;
    }

    public boolean isAllRowsEnabled()
    {
        return allRowsEnabled;
    }

    public void setAllRowsEnabled( boolean allRowsEnabled )
    {
        this.allRowsEnabled = allRowsEnabled;
    }

    public boolean isLazyLoadingEnabled()
    {
        return !isAllRowsEnabled();
    }

    public void setLazyLoadingEnabled( boolean lazyLoadingEnabled )
    {
        setAllRowsEnabled( !lazyLoadingEnabled );
    }

    public boolean isFixedTableHeightEnabled()
    {
        return fixedTableHeightEnabled;
    }

    public void setFixedTableHeightEnabled( boolean fixedTableHeightEnabled )
    {
        this.fixedTableHeightEnabled = fixedTableHeightEnabled;
    }

    public boolean isPageableTableEnabled()
    {
        return pageableTableEnabled;
    }

    public void setPageableTableEnabled( boolean pageableTableEnabled )
    {
        this.pageableTableEnabled = pageableTableEnabled;
    }

    public boolean isScrollableTableEnabled()
    {
        return !pageableTableEnabled;
    }

    public void setScrollableTableEnabled( boolean scrollableTableEnabled )
    {
        this.pageableTableEnabled = !scrollableTableEnabled;
    }

    public boolean isHorizontalLinesEnabled()
    {
        return horizontalLinesEnabled;
    }

    public void setHorizontalLinesEnabled( boolean horizontalLinesEnabled )
    {
        this.horizontalLinesEnabled = horizontalLinesEnabled;
    }

    public boolean isHighlightActiveRowEnabled()
    {
        return highlightActiveRowEnabled;
    }

    public void setHighlightActiveRowEnabled( boolean highlightActiveRowEnabled )
    {
        this.highlightActiveRowEnabled = highlightActiveRowEnabled;
    }

    public boolean isOddAndEvenColoringRowsEnabled()
    {
        return oddAndEvenColoringRowsEnabled;
    }

    public void setOddAndEvenColoringRowsEnabled( boolean oddAndEvenColoringRowsEnabled )
    {
        this.oddAndEvenColoringRowsEnabled = oddAndEvenColoringRowsEnabled;
    }

    public boolean isQuickSearchEnabled()
    {
        return quickSearchEnabled;
    }

    public void setQuickSearchEnabled( boolean quickSearchEnabled )
    {
        this.quickSearchEnabled = quickSearchEnabled;
    }

    public boolean isQuickSearchFormattingEnabled()
    {
        return quickSearchFormattingEnabled;
    }

    public void setQuickSearchFormattingEnabled( boolean quickSearchFormattingEnabled )
    {
        this.quickSearchFormattingEnabled = quickSearchFormattingEnabled;
    }

    public boolean isQuickSearchImmediateEnabled()
    {
        return quickSearchImmediateEnabled;
    }

    public void setQuickSearchImmediateEnabled( boolean quickSearchImmediateEnabled )
    {
        this.quickSearchImmediateEnabled = quickSearchImmediateEnabled;
    }

    public boolean isQuickSearchVisibleAtStartupEnabled()
    {
        return quickSearchVisibleAtStartupEnabled;
    }

    public void setQuickSearchVisibleAtStartupEnabled( boolean quickSearchVisibleAtStartupEnabled )
    {
        this.quickSearchVisibleAtStartupEnabled = quickSearchVisibleAtStartupEnabled;
    }

    public boolean isQuickSearchHideEnabled()
    {
        return quickSearchHideEnabled;
    }

    public void setQuickSearchHideEnabled( boolean quickSearchHideEnabled )
    {
        this.quickSearchHideEnabled = quickSearchHideEnabled;
    }

    public boolean isQuickSearchClearOnNoNewDataEnabled()
    {
        return quickSearchClearOnNoNewDataEnabled;
    }

    public void setQuickSearchClearOnNoNewDataEnabled( boolean quickSearchClearOnNoNewDataEnabled )
    {
        this.quickSearchClearOnNoNewDataEnabled = quickSearchClearOnNoNewDataEnabled;
    }

    public boolean isSearchEnabled()
    {
        return searchEnabled;
    }

    public void setSearchEnabled( boolean searchEnabled )
    {
        this.searchEnabled = searchEnabled;
    }

    public boolean isSearchFormattingEnabled()
    {
        return searchFormattingEnabled;
    }

    public void setSearchFormattingEnabled( boolean searchFormattingEnabled )
    {
        this.searchFormattingEnabled = searchFormattingEnabled;
    }

    public boolean isSearchImmediateEnabled()
    {
        return searchImmediateEnabled;
    }

    public void setSearchImmediateEnabled( boolean searchImmediateEnabled )
    {
        this.searchImmediateEnabled = searchImmediateEnabled;
    }

    public boolean isSearchVisibleAtStartupEnabled()
    {
        return searchVisibleAtStartupEnabled;
    }

    public void setSearchVisibleAtStartupEnabled( boolean searchVisibleAtStartupEnabled )
    {
        this.searchVisibleAtStartupEnabled = searchVisibleAtStartupEnabled;
    }

    public boolean isSearchHideEnabled()
    {
        return searchHideEnabled;
    }

    public void setSearchHideEnabled( boolean searchHideEnabled )
    {
        this.searchHideEnabled = searchHideEnabled;
    }

    public boolean isSearchClearOnNoNewDataEnabled()
    {
        return searchClearOnNoNewDataEnabled;
    }

    public void setSearchClearOnNoNewDataEnabled( boolean searchClearOnNoNewDataEnabled )
    {
        this.searchClearOnNoNewDataEnabled = searchClearOnNoNewDataEnabled;
    }

    public boolean isNoNewDataMessageEnabled()
    {
        return noNewDataMessageEnabled;
    }

    public void setNoNewDataMessageEnabled( boolean noNewDataMessageEnabled )
    {
        this.noNewDataMessageEnabled = noNewDataMessageEnabled;
    }

    public boolean isLoadingMessageEnabled()
    {
        return loadingMessageEnabled;
    }

    public void setLoadingMessageEnabled( boolean loadingMessageEnabled )
    {
        this.loadingMessageEnabled = loadingMessageEnabled;
    }

    public boolean isColumnNumberInLoadingMessageBoxEnabled()
    {
        return columnNumberInLoadingMessageBoxEnabled;
    }

    public void setColumnNumberInLoadingMessageBoxEnabled( boolean columnNumberInLoadingMessageBoxEnabled )
    {
        this.columnNumberInLoadingMessageBoxEnabled = columnNumberInLoadingMessageBoxEnabled;
    }

    public boolean isOrderedFrozenTableEnabled()
    {
        return orderedFrozenTableEnabled;
    }

    public void setOrderedFrozenTableEnabled( boolean orderedFrozenTableEnabled )
    {
        this.orderedFrozenTableEnabled = orderedFrozenTableEnabled;
    }

    public boolean isOrderedTableEnabled()
    {
        return orderedTableEnabled;
    }

    public void setOrderedTableEnabled( boolean orderedTableEnabled )
    {
        this.orderedTableEnabled = orderedTableEnabled;
    }

    public boolean isFilteredFrozenTableEnabled()
    {
        return filteredFrozenTableEnabled;
    }

    public void setFilteredFrozenTableEnabled( boolean filteredFrozenTableEnabled )
    {
        this.filteredFrozenTableEnabled = filteredFrozenTableEnabled;
    }

    public boolean isFilteredTableEnabled()
    {
        return filteredTableEnabled;
    }

    public void setFilteredTableEnabled( boolean filteredTableEnabled )
    {
        this.filteredTableEnabled = filteredTableEnabled;
    }

    public boolean isContextMenuForFilteredFrozenTableEnabled()
    {
        return contextMenuForFilteredFrozenTableEnabled;
    }

    public void setContextMenuForFilteredFrozenTableEnabled( boolean contextMenuForFilteredFrozenTableEnabled )
    {
        this.contextMenuForFilteredFrozenTableEnabled = contextMenuForFilteredFrozenTableEnabled;
    }

    public boolean isContextMenuForFilteredTableEnabled()
    {
        return contextMenuForFilteredTableEnabled;
    }

    public void setContextMenuForFilteredTableEnabled( boolean contextMenuForFilteredTableEnabled )
    {
        this.contextMenuForFilteredTableEnabled = contextMenuForFilteredTableEnabled;
    }

    public boolean isFilterInfoEnabled()
    {
        return filterInfoEnabled;
    }

    public void setFilterInfoEnabled( boolean filterInfoEnabled )
    {
        this.filterInfoEnabled = filterInfoEnabled;
    }

    public boolean isFrozenTableEnabled()
    {
        return frozenTableEnabled;
    }

    public void setFrozenTableEnabled( boolean frozenTableEnabled )
    {
        this.frozenTableEnabled = frozenTableEnabled;
    }

    public boolean isFrozenTableVisibleEnabled()
    {
        return frozenTableVisibleEnabled;
    }

    public void setFrozenTableVisibleEnabled( boolean frozenTableVisibleEnabled )
    {
        this.frozenTableVisibleEnabled = frozenTableVisibleEnabled;
    }

    public boolean isHideFrozenTableEnabled()
    {
        return hideFrozenTableEnabled;
    }

    public void setHideFrozenTableEnabled( boolean hideFrozenTableEnabled )
    {
        this.hideFrozenTableEnabled = hideFrozenTableEnabled;
    }

    public boolean isEmptyFrozenTableEnabled()
    {
        return emptyFrozenTableEnabled;
    }

    public void setEmptyFrozenTableEnabled( boolean emptyFrozenTableEnabled )
    {
        this.emptyFrozenTableEnabled = emptyFrozenTableEnabled;
    }

    public boolean isResizeFrozenTableEnabled()
    {
        return resizeFrozenTableEnabled;
    }

    public void setResizeFrozenTableEnabled( boolean resizeFrozenTableEnabled )
    {
        this.resizeFrozenTableEnabled = resizeFrozenTableEnabled;
    }

    public boolean isFreezeUnfreezeColumnsEnabled()
    {
        return freezeUnfreezeColumnsEnabled;
    }

    public void setFreezeUnfreezeColumnsEnabled( boolean freezeUnfreezeColumnsEnabled )
    {
        this.freezeUnfreezeColumnsEnabled = freezeUnfreezeColumnsEnabled;
    }

    public boolean isFreezeUnfreezeColumnsToolEnabled()
    {
        return freezeUnfreezeColumnsToolEnabled;
    }

    public void setFreezeUnfreezeColumnsToolEnabled( boolean freezeUnfreezeColumnsToolEnabled )
    {
        this.freezeUnfreezeColumnsToolEnabled = freezeUnfreezeColumnsToolEnabled;
    }

    public boolean isFreezeAnyColumnsEnabled()
    {
        return freezeAnyColumnsEnabled;
    }

    public void setFreezeAnyColumnsEnabled( boolean freezeAnyColumnsEnabled )
    {
        this.freezeAnyColumnsEnabled = freezeAnyColumnsEnabled;
    }

    public boolean isUnfreezeAnyColumnsEnabled()
    {
        return unfreezeAnyColumnsEnabled;
    }

    public void setUnfreezeAnyColumnsEnabled( boolean unfreezeAnyColumnsEnabled )
    {
        this.unfreezeAnyColumnsEnabled = unfreezeAnyColumnsEnabled;
    }

    public boolean isHideUnhideColumnsEnabled()
    {
        return hideUnhideColumnsEnabled;
    }

    public void setHideUnhideColumnsEnabled( boolean hideUnhideColumnsEnabled )
    {
        this.hideUnhideColumnsEnabled = hideUnhideColumnsEnabled;
    }

    public boolean isHideUnhideColumnsToolEnabled()
    {
        return hideUnhideColumnsToolEnabled;
    }

    public void setHideUnhideColumnsToolEnabled( boolean hideUnhideColumnsToolEnabled )
    {
        this.hideUnhideColumnsToolEnabled = hideUnhideColumnsToolEnabled;
    }

    public boolean isHideAnyColumnsEnabled()
    {
        return hideAnyColumnsEnabled;
    }

    public void setHideAnyColumnsEnabled( boolean hideAnyColumnsEnabled )
    {
        this.hideAnyColumnsEnabled = hideAnyColumnsEnabled;
    }

    public boolean isUnhideAnyColumnsEnabled()
    {
        return unhideAnyColumnsEnabled;
    }

    public void setUnhideAnyColumnsEnabled( boolean unhideAnyColumnsEnabled )
    {
        this.unhideAnyColumnsEnabled = unhideAnyColumnsEnabled;
    }

    public boolean isNewDataWhenHideUnhideColumnEnabled()
    {
        return newDataWhenHideUnhideColumnEnabled;
    }

    public void setNewDataWhenHideUnhideColumnEnabled( boolean newDataWhenHideUnhideColumnEnabled )
    {
        this.newDataWhenHideUnhideColumnEnabled = newDataWhenHideUnhideColumnEnabled;
    }

    public boolean isTableToolsMenuEnabled()
    {
        return tableToolsMenuEnabled;
    }

    public void setTableToolsMenuEnabled( boolean tableToolsMenuEnabled )
    {
        this.tableToolsMenuEnabled = tableToolsMenuEnabled;
    }

    public boolean isTableToolsMenuAlwaysVisibleEnabled()
    {
        return tableToolsMenuAlwaysVisibleEnabled;
    }

    public void setTableToolsMenuAlwaysVisibleEnabled( boolean tableToolsMenuAlwaysVisibleEnabled )
    {
        this.tableToolsMenuAlwaysVisibleEnabled = tableToolsMenuAlwaysVisibleEnabled;
    }

    public boolean isTableToolsTrayEnabled()
    {
        return tableToolsTrayEnabled;
    }

    public void setTableToolsTrayEnabled( boolean tableToolsTrayEnabled )
    {
        this.tableToolsTrayEnabled = tableToolsTrayEnabled;
    }

    public float getLoadingMessageOpacity()
    {
        return loadingMessageOpacity;
    }

    public void setLoadingMessageOpacity( float loadingMessageOpacity )
    {
        this.loadingMessageOpacity =
            loadingMessageOpacity > 0 ? loadingMessageOpacity : DEFAULT_LOADING_MESSAGE_OPACITY;
    }

    public void setLoadingMessageDelay( int showDelay, int hideDelay )
    {
        if ( DEFAULT_MINIMAL_DELAY <= showDelay && DEFAULT_MINIMAL_DELAY <= hideDelay )
        {
            loadingMessageShowDelay = showDelay;
            loadingMessageHideDelay = hideDelay;
        }
    }

    public int getLoadingMessageShowDelay()
    {
        return loadingMessageShowDelay;
    }

    public void setLoadingMessageShowDelay( int loadingMessageShowDelay )
    {
        this.loadingMessageShowDelay = DEFAULT_MINIMAL_DELAY <= loadingMessageShowDelay ? loadingMessageShowDelay
                        : DEFAULT_LOADING_MESSAGE_SHOW_DELAY;
    }

    public int getLoadingMessageHideDelay()
    {
        return loadingMessageHideDelay;
    }

    public void setLoadingMessageHideDelay( int loadingMessageHideDelay )
    {
        this.loadingMessageHideDelay = DEFAULT_MINIMAL_DELAY <= loadingMessageHideDelay ? loadingMessageHideDelay
                        : DEFAULT_LOADING_MESSAGE_HIDE_DELAY;
    }

    private void checkColumnsFreezeEnabled()
    {
        // check column number
        for ( int column = columnFreezeEnabled.size() - 1; 0 <= column; column-- )
        {
            int col = Integer.parseInt( columnFreezeEnabled.get( column )
                                                           .substring( 1 ) );
            if ( !( 0 <= col && col < allColumns ) )
            {
                columnFreezeEnabled.remove( column );
            }
        }

        // removed duplicated values
        if ( !columnFreezeEnabled.isEmpty() )
        {
            Set<String> tempColumns = new LinkedHashSet<>( columnFreezeEnabled );
            columnFreezeEnabled = new ArrayList<>( tempColumns );
        }
    }

    @Override
    public boolean canColumnFreeze( int column )
    {
        if ( columnFreezeEnabled.contains( "-" + column ) )
        {
            return false;
        }
        else if ( columnFreezeEnabled.contains( "+" + column ) )
        {
            return true;
        }
        return freezeAnyColumnsEnabled;
    }

    @Override
    public void setColumnFreezeEnabled( boolean columnFreezeEnabled, int column )
    {
        try
        {
            this.columnFreezeEnabled.remove( "+" + column );
            this.columnFreezeEnabled.remove( "-" + column );
        }
        catch ( Exception e )
        {
            Logger.getLogger( DTable.class.getName() )
                  .log( Level.INFO, e.getLocalizedMessage(), e );
        }
        this.columnFreezeEnabled.add( ( columnFreezeEnabled ? "+" : "-" ) + column );
    }

    @Override
    public boolean isColumnsFreezeEmptyEnabled()
    {
        return !( freezeAnyColumnsEnabled && !columnFreezeEnabled.isEmpty() );
    }

    @Override
    public String[] getColumnsFreezeEnabled( int[] columns )
    {
        if ( freezeAnyColumnsEnabled || !columnFreezeEnabled.isEmpty() )
        {
            checkColumnsFreezeEnabled();
            List<Integer> cols = Arrays.asList( ArrayUtils.toObject( columns ) );
            List<String> tempCols = new ArrayList<>();
            for ( String column : columnFreezeEnabled )
            {
                if ( cols.contains( Integer.parseInt( column.substring( 1 ) ) ) )
                {
                    tempCols.add( column );
                }
            }
            return tempCols.toArray( new String[] {} );
        }
        return new String[] {};
    }

    @Override
    public String[] getColumnsFreezeEnabled()
    {
        if ( freezeAnyColumnsEnabled && !columnFreezeEnabled.isEmpty() )
        {
            checkColumnsFreezeEnabled();
            return columnFreezeEnabled.toArray( new String[] {} );
        }
        return new String[] {};
    }

    @Override
    public void setColumnsFreezeEnabled( boolean columnsFreezeEnabled, int... columns )
    {
        for ( int column : columns )
        {
            setColumnFreezeEnabled( columnsFreezeEnabled, column );
        }
    }

    private void checkColumnsUnfreezeEnabled()
    {
        // check column number
        for ( int column = columnUnfreezeEnabled.size() - 1; 0 <= column; column-- )
        {
            int col = Integer.parseInt( columnUnfreezeEnabled.get( column )
                                                             .substring( 1 ) );
            if ( !( 0 <= col && col < allColumns ) )
            {
                columnUnfreezeEnabled.remove( column );
            }
        }

        // removed duplicated values
        if ( !columnUnfreezeEnabled.isEmpty() )
        {
            Set<String> tempColumns = new LinkedHashSet<>( columnUnfreezeEnabled );
            columnUnfreezeEnabled = new ArrayList<>( tempColumns );
        }
    }

    @Override
    public boolean canColumnUnfreeze( int column )
    {
        if ( columnUnfreezeEnabled.contains( "-" + column ) )
        {
            return false;
        }
        else if ( columnUnfreezeEnabled.contains( "+" + column ) )
        {
            return true;
        }
        return unfreezeAnyColumnsEnabled;
    }

    @Override
    public void setColumnUnfreezeEnabled( boolean columnUnfreezeEnabled, int column )
    {
        try
        {
            this.columnUnfreezeEnabled.remove( "+" + column );
            this.columnUnfreezeEnabled.remove( "-" + column );
        }
        catch ( Exception e )
        {
            Logger.getLogger( DTable.class.getName() )
                  .log( Level.INFO, e.getLocalizedMessage(), e );
        }
        this.columnUnfreezeEnabled.add( ( columnUnfreezeEnabled ? "+" : "-" ) + column );
    }

    @Override
    public boolean isColumnsUnfreezeEmptyEnabled()
    {
        return !( unfreezeAnyColumnsEnabled && !columnUnfreezeEnabled.isEmpty() );
    }

    @Override
    public String[] getColumnsUnfreezeEnabled( int[] columns )
    {
        if ( unfreezeAnyColumnsEnabled || !columnUnfreezeEnabled.isEmpty() )
        {
            checkColumnsUnfreezeEnabled();
            List<Integer> cols = Arrays.asList( ArrayUtils.toObject( columns ) );
            List<String> tempCols = new ArrayList<>();
            for ( String column : columnUnfreezeEnabled )
            {
                if ( cols.contains( Integer.parseInt( column.substring( 1 ) ) ) )
                {
                    tempCols.add( column );
                }
            }
            return tempCols.toArray( new String[] {} );
        }
        return new String[] {};
    }

    @Override
    public String[] getColumnsUnfreezeEnabled()
    {
        if ( unfreezeAnyColumnsEnabled && !columnUnfreezeEnabled.isEmpty() )
        {
            checkColumnsUnfreezeEnabled();
            return columnUnfreezeEnabled.toArray( new String[] {} );
        }
        return new String[] {};
    }

    @Override
    public void setColumnsUnfreezeEnabled( boolean columnsUnfreezeEnabled, int... columns )
    {
        for ( int column : columns )
        {
            setColumnUnfreezeEnabled( columnsUnfreezeEnabled, column );
        }
    }

    private void checkColumnsHideEnabled()
    {
        // check column number
        for ( int column = columnHideEnabled.size() - 1; 0 <= column; column-- )
        {
            int col = Integer.parseInt( columnHideEnabled.get( column )
                                                         .substring( 1 ) );
            if ( !( 0 <= col && col < allColumns ) )
            {
                columnHideEnabled.remove( column );
            }
        }

        // removed duplicated values
        if ( !columnHideEnabled.isEmpty() )
        {
            Set<String> tempColumns = new LinkedHashSet<>( columnHideEnabled );
            columnHideEnabled = new ArrayList<>( tempColumns );
        }
    }

    @Override
    public boolean canColumnHide( int column )
    {
        if ( columnHideEnabled.contains( "-" + column ) )
        {
            return false;
        }
        else if ( columnHideEnabled.contains( "+" + column ) )
        {
            return true;
        }
        return hideAnyColumnsEnabled;
    }

    @Override
    public void setColumnHideEnabled( boolean columnHideEnabled, int column )
    {
        try
        {
            this.columnHideEnabled.remove( "+" + column );
            this.columnHideEnabled.remove( "-" + column );
        }
        catch ( Exception e )
        {
            Logger.getLogger( DTable.class.getName() )
                  .log( Level.INFO, e.getLocalizedMessage(), e );
        }
        this.columnHideEnabled.add( ( columnHideEnabled ? "+" : "-" ) + column );
    }

    @Override
    public String[] getColumnsHideEnabled( int[] columns )
    {
        if ( hideAnyColumnsEnabled || !columnHideEnabled.isEmpty() )
        {
            checkColumnsHideEnabled();
            List<Integer> cols = Arrays.asList( ArrayUtils.toObject( columns ) );
            List<String> tempCols = new ArrayList<>();
            for ( String column : columnHideEnabled )
            {
                if ( cols.contains( Integer.parseInt( column.substring( 1 ) ) ) )
                {
                    tempCols.add( column );
                }
            }
            return tempCols.toArray( new String[] {} );
        }
        return new String[] {};
    }

    @Override
    public String[] getColumnsHideEnabled()
    {
        if ( hideAnyColumnsEnabled && !columnHideEnabled.isEmpty() )
        {
            checkColumnsHideEnabled();
            return columnHideEnabled.toArray( new String[] {} );
        }
        return new String[] {};
    }

    @Override
    public void setColumnsHideEnabled( boolean columnsHideEnabled, int... columns )
    {
        for ( int column : columns )
        {
            setColumnHideEnabled( columnsHideEnabled, column );
        }
    }

    private void checkColumnsUnhideEnabled()
    {
        // check column number
        for ( int column = columnUnhideEnabled.size() - 1; 0 <= column; column-- )
        {
            int col = Integer.parseInt( columnUnhideEnabled.get( column )
                                                           .substring( 1 ) );
            if ( !( 0 <= col && col < allColumns ) )
            {
                columnUnhideEnabled.remove( column );
            }
        }

        // removed duplicated values
        if ( !columnUnhideEnabled.isEmpty() )
        {
            Set<String> tempColumns = new LinkedHashSet<>( columnUnhideEnabled );
            columnUnhideEnabled = new ArrayList<>( tempColumns );
        }
    }

    @Override
    public boolean canColumnUnhide( int column )
    {
        if ( columnUnhideEnabled.contains( "-" + column ) )
        {
            return false;
        }
        else if ( columnUnhideEnabled.contains( "+" + column ) )
        {
            return true;
        }
        return unhideAnyColumnsEnabled;
    }

    @Override
    public void setColumnUnhideEnabled( boolean columnUnhideEnabled, int column )
    {
        try
        {
            this.columnUnhideEnabled.remove( "+" + column );
            this.columnUnhideEnabled.remove( "-" + column );
        }
        catch ( Exception e )
        {
            Logger.getLogger( DTable.class.getName() )
                  .log( Level.INFO, e.getLocalizedMessage(), e );
        }
        this.columnUnhideEnabled.add( ( columnUnhideEnabled ? "+" : "-" ) + column );
    }

    @Override
    public String[] getColumnsUnhideEnabled( int[] columns )
    {
        if ( unhideAnyColumnsEnabled || !columnUnhideEnabled.isEmpty() )
        {
            checkColumnsUnhideEnabled();
            List<Integer> cols = Arrays.asList( ArrayUtils.toObject( columns ) );
            List<String> tempCols = new ArrayList<>();
            for ( String column : columnUnhideEnabled )
            {
                if ( cols.contains( Integer.parseInt( column.substring( 1 ) ) ) )
                {
                    tempCols.add( column );
                }
            }
            return tempCols.toArray( new String[] {} );
        }
        return new String[] {};
    }

    @Override
    public String[] getColumnsUnhideEnabled()
    {
        if ( unhideAnyColumnsEnabled && !columnUnhideEnabled.isEmpty() )
        {
            checkColumnsUnhideEnabled();
            return columnUnhideEnabled.toArray( new String[] {} );
        }
        return new String[] {};
    }

    @Override
    public void setColumnsUnhideEnabled( boolean columnsUnhideEnabled, int... columns )
    {
        for ( int column : columns )
        {
            setColumnUnhideEnabled( columnsUnhideEnabled, column );
        }
    }

    private void checkFormattingColumns()
    {
        if ( !isNoFormattingColumn() )
        {
            for ( int column = formattingColumns.size() - 1; 0 <= column; column-- )
            {
                if ( !( 0 <= formattingColumns.get( column )
                                              .getColumn()
                    && formattingColumns.get( column )
                                        .getColumn() < allColumns ) )
                {
                    formattingColumns.remove( column );
                }
            }
        }
    }

    private void replaceFormattingColumn( Formatting newFormattingColumns )
    {
        if ( !isNoFormattingColumn() )
        {
            removeFormattingColumn( newFormattingColumns.getColumn() );
        }
        formattingColumns.add( newFormattingColumns );
    }

    @Override
    public void removeFormattingColumn( int column )
    {
        if ( !isNoFormattingColumn() && isFormattingColumn( column ) )
        {
            for ( int col = 0; col < formattingColumns.size(); col++ )
            {
                if ( formattingColumns.get( col )
                                      .equals( column ) )
                {
                    formattingColumns.remove( col );
                    break;
                }
            }
        }
    }

    @Override
    public boolean isNoFormattingColumn()
    {
        return !( formattingColumns != null && !formattingColumns.isEmpty() );
    }

    @Override
    public boolean isFormattingColumn( int column )
    {
        if ( !isNoFormattingColumn() )
        {
            for ( Formatting formatting : formattingColumns )
            {
                if ( formatting.getColumn()
                               .equals( column ) )
                {
                    return true;
                }
            }
        }
        return false;
    }

    @Override
    public Formatting getFormattingColumn( int column )
    {
        if ( !isNoFormattingColumn() )
        {
            checkFormattingColumns();
            for ( Formatting formatting : formattingColumns )
            {
                if ( formatting.getColumn()
                               .equals( column ) )
                {
                    return formatting;
                }
            }
        }
        return null;
    }

    @Override
    public void setFormattingColumn( int column, Format format )
    {
        if ( format != null )
        {
            replaceFormattingColumn( new Formatting( column, format ) );
        }
    }

    @Override
    public Formatting[] getFormattingColumnsArray()
    {
        if ( !isNoFormattingColumn() )
        {
            checkFormattingColumns();
            return formattingColumns.toArray( new Formatting[] {} );
        }
        return new Formatting[] {};
    }

    @Override
    public int[] getFormattingColumns( int[] columns )
    {
        if ( !isNoFormattingColumn() || columns == null || columns.length == 0 )
        {
            checkFormattingColumns();
            List<Integer> cols = Arrays.asList( ArrayUtils.toObject( columns ) );
            List<Integer> tempCols = new ArrayList<>();
            for ( Formatting formatting : formattingColumns )
            {
                if ( cols.contains( formatting.getColumn() ) )
                {
                    tempCols.add( formatting.getColumn() );
                }
            }
            return ArrayUtils.toPrimitive( tempCols.toArray( new Integer[] {} ) );
        }
        return new int[] {};
    }

    @Override
    public int[] getFormattingColumns()
    {
        if ( !isNoFormattingColumn() )
        {
            List<Integer> tempCols = new ArrayList<>();
            for ( Formatting formatting : formattingColumns )
            {
                tempCols.add( formatting.getColumn() );
            }
            return ArrayUtils.toPrimitive( tempCols.toArray( new Integer[] {} ) );
        }
        return new int[] {};
    }

    @Override
    public void setFormattingColumns( Format format, int... columns )
    {
        if ( format != null && columns != null && 0 < columns.length )
        {
            for ( int column : columns )
            {
                setFormattingColumn( column, format );
            }
        }
    }

    @Override
    public int[] getDateOrNumericFormattingColumns()
    {
        if ( !isNoFormattingColumn() )
        {
            List<Integer> tempCols = new ArrayList<>();
            for ( Formatting formatting : formattingColumns )
            {
                switch ( getColumnType( formatting.getColumn() ) )
                {
                    case BIG_DECIMAL:
                    case BIG_INTEGER:
                    case DATE:
                    case DOUBLE:
                    case FLOAT:
                    case INTEGER:
                    case LONG:
                    case BOOLEAN:
                        tempCols.add( formatting.getColumn() );
                        break;
                    case MEMO:
                    case TEXT:
                    default:
                        break;
                }

            }
            return ArrayUtils.toPrimitive( tempCols.toArray( new Integer[] {} ) );
        }
        return new int[] {};
    }

    @Override
    public int[] getDateOrNumericFormattingColumns( int[] columns )
    {
        if ( !isNoFormattingColumn() || columns == null || columns.length == 0 )
        {
            checkFormattingColumns();
            List<Integer> cols = Arrays.asList( ArrayUtils.toObject( columns ) );
            List<Integer> tempCols = new ArrayList<>();
            for ( Formatting formatting : formattingColumns )
            {
                if ( cols.contains( formatting.getColumn() ) )
                {
                    switch ( getColumnType( formatting.getColumn() ) )
                    {
                        case BIG_DECIMAL:
                        case BIG_INTEGER:
                        case DATE:
                        case DOUBLE:
                        case FLOAT:
                        case INTEGER:
                        case LONG:
                        case BOOLEAN:
                            tempCols.add( formatting.getColumn() );
                            break;
                        case MEMO:
                        case TEXT:
                        default:
                            break;

                    }
                }
            }
            return ArrayUtils.toPrimitive( tempCols.toArray( new Integer[] {} ) );
        }
        return new int[] {};
    }

    public Map<String, String> getData()
    {
        Map<String, String> data = new HashMap<>();

        data.put( "pageResizingActionDelay", Integer.toString( getPageResizingActionDelay() ) );
        data.put( "scrollbarActionDelay", Integer.toString( getScrollbarActionDelay() ) );
        data.put( "tableToolsMouseClickActionDelay", Integer.toString( getTableToolsMouseClickActionDelay() ) );
        data.put( "tableToolsTrayMouseOverAnimationDuration",
                  Integer.toString( getTableToolsTrayMouseOverAnimationDuration() ) );
        data.put( "tableToolsTrayAnimationDuration", Integer.toString( getTableToolsTrayAnimationDuration() ) );
        data.put( "quickSearchAnimationDuration", Integer.toString( getQuickSearchAnimationDuration() ) );
        data.put( "searchAnimationDuration", Integer.toString( getSearchAnimationDuration() ) );
        data.put( "pageSize", Integer.toString( getPageSize() ) );
        data.put( "pagerPages", Integer.toString( getPagerPages() ) );
        data.put( "tableToolsPosition", getTableToolsPosition() );
        data.put( "filterInfoPosition", getFilterInfoPosition() );
        data.put( "tableFrozenMinimalWidth", Integer.toString( getTableFrozenMinimalWidth() ) );
        data.put( "tableMinimalWidth", Integer.toString( getTableMinimalWidth() ) );
        data.put( "contextMenuDepthForFilteredFrozenTable",
                  Integer.toString( getContextMenuDepthForFilteredFrozenTable() ) );
        data.put( "contextMenuDepthForFilteredTable", Integer.toString( getContextMenuDepthForFilteredTable() ) );
        data.put( "tableHeight", Integer.toString( getTableHeight() ) );
        data.put( "fixedTableHeightEnabled", Boolean.toString( isFixedTableHeightEnabled() ) );

        data.put( "headerRowHeight", Integer.toString( getHeader().getRowHeight() ) );
        data.put( "headerLineHeight", Integer.toString( getHeader().getLineHeight() ) );
        data.put( "headerCellMarginTop", Integer.toString( getHeader().getTopMargin() ) );
        data.put( "headerCellMarginRight", Integer.toString( getHeader().getRightMargin() ) );
        data.put( "headerCellMarginBottom", Integer.toString( getHeader().getBottomMargin() ) );
        data.put( "headerCellMarginLeft", Integer.toString( getHeader().getLeftMargin() ) );
        data.put( "headerRowsAlignment", Integer.toString( getHeader().getRowAlignment()
                                                                      .getAlignment() ) );
        data.put( "allHeaderColumnsEnabled", Boolean.toString( getHeader().isAllColumnsEnabled() ) );
        data.put( "headerVerticalLinesEnabled", Boolean.toString( getHeader().isVerticalLinesEnabled() ) );
        data.put( "expandedFrozenHeaderCellEnabled", Boolean.toString( getHeader().isExpandedFrozenCellEnabled() ) );
        data.put( "expandedFrozenHeaderCellAlwaysVisibleEnabled",
                  Boolean.toString( getHeader().isExpandedFrozenCellAlwaysVisibleEnabled() ) );
        data.put( "expandedFrozenHeaderCellForActiveColumnEnabled",
                  Boolean.toString( getHeader().isExpandedFrozenCellForActiveColumnEnabled() ) );
        data.put( "expandedHeaderCellEnabled", Boolean.toString( getHeader().isExpandedCellEnabled() ) );
        data.put( "expandedHeaderCellAlwaysVisibleEnabled",
                  Boolean.toString( getHeader().isExpandedCellAlwaysVisibleEnabled() ) );
        data.put( "expandedHeaderCellForActiveColumnEnabled",
                  Boolean.toString( getHeader().isExpandedCellForActiveColumnEnabled() ) );
        data.put( "wordWrapInHeaderCellsEnabled", Boolean.toString( getHeader().isWordWrapEnabled() ) );
        data.put( "tooltipEnabled", Boolean.toString( getHeader().isTooltipEnabled() ) );
        data.put( "wordWrapInTooltipEnabled", Boolean.toString( getHeader().isWordWrapInTooltipEnabled() ) );
        data.put( "tooltipPosition", getHeader().getTooltipPosition()
                                                .toString() );
        data.put( "tooltipMinimumWidth", Integer.toString( getHeader().getTooltipMinimumWidth() ) );
        data.put( "tooltipMaximumWidth", Integer.toString( getHeader().getTooltipMaximumWidth() ) );

        data.put( "bodyRowHeight", Integer.toString( getBody().getRowHeight() ) );
        data.put( "bodyLineHeight", Integer.toString( getBody().getLineHeight() ) );
        data.put( "bodyCellMarginTop", Integer.toString( getBody().getTopMargin() ) );
        data.put( "bodyCellMarginRight", Integer.toString( getBody().getRightMargin() ) );
        data.put( "bodyCellMarginBottom", Integer.toString( getBody().getBottomMargin() ) );
        data.put( "bodyCellMarginLeft", Integer.toString( getBody().getLeftMargin() ) );
        data.put( "bodyRowsAlignment", Integer.toString( getBody().getRowAlignment()
                                                                  .getAlignment() ) );
        data.put( "bodyVerticalLinesEnabled", Boolean.toString( getBody().isVerticalLinesEnabled() ) );
        data.put( "expandedFrozenBodyCellEnabled", Boolean.toString( getBody().isExpandedFrozenCellEnabled() ) );
        data.put( "expandedFrozenBodyCellAlwaysVisibleEnabled",
                  Boolean.toString( getBody().isExpandedFrozenCellAlwaysVisibleEnabled() ) );
        data.put( "expandedBodyCellEnabled", Boolean.toString( getBody().isExpandedCellEnabled() ) );
        data.put( "expandedBodyCellAlwaysVisibleEnabled",
                  Boolean.toString( getBody().isExpandedCellAlwaysVisibleEnabled() ) );
        data.put( "wordWrapInBodyCellsEnabled", Boolean.toString( getBody().isWordWrapEnabled() ) );

        data.put( "allRowsEnabled", Boolean.toString( isAllRowsEnabled() ) );
        data.put( "allColumnsEnabled", Boolean.toString( isAllColumnsEnabled() ) );
        data.put( "quickSearchEnabled", Boolean.toString( isQuickSearchEnabled() ) );
        data.put( "quickSearchImmediateEnabled", Boolean.toString( isQuickSearchImmediateEnabled() ) );
        data.put( "quickSearchVisibleAtStartupEnabled", Boolean.toString( isQuickSearchVisibleAtStartupEnabled() ) );
        data.put( "quickSearchHideEnabled", Boolean.toString( isQuickSearchHideEnabled() ) );
        data.put( "searchEnabled", Boolean.toString( isSearchEnabled() ) );
        data.put( "searchFormattingEnabled", Boolean.toString( isSearchFormattingEnabled() ) );
        data.put( "searchImmediateEnabled", Boolean.toString( isSearchImmediateEnabled() ) );
        data.put( "searchVisibleAtStartupEnabled", Boolean.toString( isSearchVisibleAtStartupEnabled() ) );
        data.put( "searchHideEnabled", Boolean.toString( isSearchHideEnabled() ) );
        data.put( "columnNumberInWaitBoxEnabled", Boolean.toString( isColumnNumberInLoadingMessageBoxEnabled() ) );
        data.put( "frozenTableEnabled", Boolean.toString( isFrozenTableEnabled() ) );
        data.put( "frozenTableVisibleEnabled", Boolean.toString( isFrozenTableVisibleEnabled() ) );
        data.put( "hideFrozenTableEnabled", Boolean.toString( isHideFrozenTableEnabled() ) );
        data.put( "filteredFrozenTableEnabled", Boolean.toString( isFilteredFrozenTableEnabled() ) );
        data.put( "filteredTableEnabled", Boolean.toString( isFilteredTableEnabled() ) );
        data.put( "contextMenuForFilteredFrozenTableEnabled",
                  Boolean.toString( isContextMenuForFilteredFrozenTableEnabled() ) );
        data.put( "contextMenuForFilteredTableEnabled", Boolean.toString( isContextMenuForFilteredTableEnabled() ) );
        data.put( "emptyFrozenTableEnabled", Boolean.toString( isEmptyFrozenTableEnabled() ) );
        data.put( "resizeFrozenTableEnabled", Boolean.toString( isResizeFrozenTableEnabled() ) );
        data.put( "orderedFrozenTableEnabled", Boolean.toString( isOrderedFrozenTableEnabled() ) );
        data.put( "orderedTableEnabled", Boolean.toString( isOrderedTableEnabled() ) );
        data.put( "filterInfoEnabled", Boolean.toString( isFilterInfoEnabled() ) );
        data.put( "pageableTableEnabled", Boolean.toString( isPageableTableEnabled() ) );
        data.put( "freezeUnfreezeColumnsEnabled", Boolean.toString( isFreezeUnfreezeColumnsEnabled() ) );
        data.put( "freezeUnfreezeColumnsToolEnabled", Boolean.toString( isFreezeUnfreezeColumnsToolEnabled() ) );
        data.put( "freezeAnyColumnsEnabled", Boolean.toString( isFreezeAnyColumnsEnabled() ) );
        data.put( "unfreezeAnyColumnsEnabled", Boolean.toString( isUnfreezeAnyColumnsEnabled() ) );
        data.put( "hideUnhideColumnsEnabled", Boolean.toString( isHideUnhideColumnsEnabled() ) );
        data.put( "hideUnhideColumnsToolEnabled", Boolean.toString( isHideUnhideColumnsToolEnabled() ) );
        data.put( "hideAnyColumnsEnabled", Boolean.toString( isHideAnyColumnsEnabled() ) );
        data.put( "newDataWhenHideUnhideColumnEnabled", Boolean.toString( isNewDataWhenHideUnhideColumnEnabled() ) );
        data.put( "horizontalLinesEnabled", Boolean.toString( isHorizontalLinesEnabled() ) );
        data.put( "highlightActiveRowEnabled", Boolean.toString( isHighlightActiveRowEnabled() ) );
        data.put( "oddAndEvenColoringRowsEnabled", Boolean.toString( isOddAndEvenColoringRowsEnabled() ) );
        data.put( "loadingMessageEnabled", Boolean.toString( isLoadingMessageEnabled() ) );
        data.put( "loadingMessageOpacity", "" + "" + getLoadingMessageOpacity() );
        data.put( "loadingMessageShowDelay", Integer.toString( getLoadingMessageShowDelay() ) );
        data.put( "loadingMessageHideDelay", Integer.toString( getLoadingMessageHideDelay() ) );
        data.put( "tableToolsMenuEnabled", Boolean.toString( isTableToolsMenuEnabled() ) );
        data.put( "tableToolsMenuAlwaysVisibleEnabled", Boolean.toString( isTableToolsMenuAlwaysVisibleEnabled() ) );
        data.put( "tableToolsTrayEnabled", Boolean.toString( isTableToolsTrayEnabled() ) );

        return data;
    }

    public Map<String, String> getData( int[] columns, int[] headerColumns )
    {
        Map<String, String> data = new HashMap<>();

        data.put( "headerColumnsAlignment",
                  Arrays.toString( getHeader().getColumnsAlignmentIntArray( headerColumns ) ) );
        data.put( "headerColumnsWidth", Arrays.toString( getColumnsWidth( headerColumns ) ) );
        data.put( "bodyColumnsAlignment", Arrays.toString( getBody().getColumnsAlignmentIntArray( columns ) ) );
        data.put( "columnsWidth", Arrays.toString( getColumnsWidth( columns ) ) );
        data.put( "formatedColumns", Arrays.toString( getDateOrNumericFormattingColumns( columns ) ) );
        data.put( "freezeColumnsEnabled", Arrays.toString( getColumnsFreezeEnabled( columns ) ) );
        data.put( "unfreezeColumnsEnabled", Arrays.toString( getColumnsUnfreezeEnabled( columns ) ) );
        data.put( "hideColumnsEnabled", Arrays.toString( getColumnsHideEnabled( columns ) ) );

        return data;
    }

}
