/**
 * Copyright (C) 2012-2017  Dušan Vejnovič  <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.dussan.vaadin.dtable.windows.exporter;

import java.io.BufferedWriter;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dtable.DTable;
import org.dussan.vaadin.dtable.translations.TranslatedWords;
import org.dussan.vaadin.dtable.windows.exporter.base.Exporter;
import org.dussan.vaadin.dtable.windows.exporter.type.ExportedType;

public class CsvExporter
    extends Exporter
{

    private static final long serialVersionUID = -7064150607674977016L;

    private BufferedWriter bufferedWriter = null;

    private FileOutputStream fileOutputStream = null;

    private OutputStreamWriter outputStreamWriter = null;

    public CsvExporter( DTable dTable )
    {
        super( dTable );
        setCaption( getDTable().getTranslation()
                               .get( getDTable().getTableConfig()
                                                .getLocale(),
                                     TranslatedWords.EXPORTING_DATA_TO_CSV_FILE ) );
        setExportedType( ExportedType.CSV );
    }

    private String createHeader( String name )
    {
        String delimiter = getDTable().getTableExporterConfig()
                                      .getCsvColumnDelimiter();
        List<Integer> columns = new ArrayList<>( getDTable().getTableExporterConfig()
                                                            .getColumnDefinition( name )
                                                            .keySet() );
        Collections.sort( columns );

        StringBuilder header = new StringBuilder();
        if ( getDTable().getTableExporterConfig()
                        .isRowIndexEnabled() )
        {
            header.append( delimiter + "\"#\"" );
        }
        for ( int column : columns )
        {
            header.append( delimiter + "\"" );
            header.append( getDTable().getTableExporterConfig()
                                      .getColumnDefinitionHeader( name, column ) );
            header.append( "\"" );
        }
        return header.substring( delimiter.length() );
    }

    private String row2csv( Object[] rowData, String delimiter )
    {
        return "\"" + StringUtils.join( rowData, "\"" + delimiter + "\"" ) + "\"";
    }

    @Override
    protected void exportDataIntoFile()
    {
        try
        {
            if ( 0 < getDTable().getTableExporterConfig()
                                .getColumnsDefinitionName().length )
            {
                float completeStep = 0f;
                float step = 1f / getDTable().getTableExporterConfig()
                                             .getColumnsDefinitionName().length;

                setExportedStatus( EXPORT_BEGIN );

                for ( String name : getDTable().getTableExporterConfig()
                                               .getColumnsDefinitionName() )
                {
                    if ( openFile( name ) && prepareExportedData( name ) && canContinue() )
                    {
                        writeToFile( name, completeStep, step );
                        closeFile();
                        completeStep += step;
                    }
                    else
                    {
                        setExportedStatus( EXPORT_ERROR );
                    }

                    // check status
                    if ( !canContinue() )
                    {
                        closeFile();
                        if ( getExportedStatus() == EXPORT_STOP )
                        {
                            setExportedStatus( EXPORT_STOPPED );
                        }
                        return;
                    }
                }

                setExportedStatus( EXPORT_COMPLETE );
            }
            else
            {
                setExportedStatus( EXPORT_ERROR, getDTable().getTranslation()
                                                            .get( getDTable().getTableConfig()
                                                                             .getLocale(),
                                                                  TranslatedWords.NO_DATA_FOR_EXPORT ) );
            }
        }
        catch ( Exception e )
        {
            Logger.getLogger( DTable.class.getName() )
                  .log( Level.INFO, e.getLocalizedMessage(), e );
            setExportedStatus( EXPORT_ERROR, e.getLocalizedMessage() );
        }
    }

    @Override
    protected boolean openFile( String name )
    {
        try
        {
            fileOutputStream = new FileOutputStream( getTempFile( name ) );
            outputStreamWriter = new OutputStreamWriter( fileOutputStream, getCharset() );
            bufferedWriter = new BufferedWriter( outputStreamWriter );
            return true;
        }
        catch ( Exception e )
        {
            Logger.getLogger( DTable.class.getName() )
                  .log( Level.INFO, e.getLocalizedMessage(), e );
            setExportedStatus( EXPORT_ERROR, e.getLocalizedMessage() );
            return false;
        }
    }

    @Override
    protected void writeToFile( String name, float completeStep, float step )
    {
        int row = 0;
        int rows = 0;
        try
        {
            long allRows = getAllRows( name );
            int allColumns = getAllColumns( name );
            String delimiter = getDTable().getTableExporterConfig()
                                          .getCsvColumnDelimiter();
            if ( 0 < allColumns && 0 < allRows && 0 < getExportedStep( name ) )
            {
                int exportedStep = allRows < getExportedStep( name ) ? (int) allRows : getExportedStep( name );

                StringBuilder exportedData = new StringBuilder();
                exportedData.append( createHeader( name ) + "\n" );

                for ( int mainRow = 0; mainRow < allRows; mainRow += exportedStep )
                {
                    rows = (int) ( mainRow + exportedStep <= allRows ? exportedStep : allRows - mainRow );
                    Object[] rowsData = getFromattedData( name, mainRow, rows );

                    for ( row = 0; row < rows; row++ )
                    {
                        if ( getDTable().getTableExporterConfig()
                                        .isRowIndexEnabled() )
                        {
                            exportedData.append( "\"" + ( mainRow + row + 1 ) + "\"" + delimiter );
                        }
                        exportedData.append( row2csv( Arrays.copyOfRange( rowsData, row * allColumns,
                                                                          ( row + 1 ) * allColumns ),
                                                      delimiter )
                            + "\n" );

                        // exports was stopped
                        if ( !canContinue() )
                        {
                            setExportedStatus( EXPORT_STOPPED );
                            return;
                        }
                    }

                    setExportedStatus( ( step * mainRow / allRows ) + completeStep );
                    bufferedWriter.write( exportedData.toString() );
                    exportedData = new StringBuilder();
                }
            }
            else
            {
                setExportedStatus( EXPORT_ERROR, getDTable().getTranslation()
                                                            .get( getDTable().getTableConfig()
                                                                             .getLocale(),
                                                                  TranslatedWords.NO_DATA_FOR_EXPORT ) );
            }
        }
        catch ( Exception e )
        {
            Logger.getLogger( DTable.class.getName() )
                  .log( Level.INFO, e.getLocalizedMessage(), e );
            setExportedStatus( EXPORT_ERROR, e.getLocalizedMessage() );
        }
    }

    @Override
    protected void closeFile()
    {
        try
        {
            bufferedWriter.close();
            outputStreamWriter.close();
            fileOutputStream.close();
        }
        catch ( Exception e )
        {
            Logger.getLogger( DTable.class.getName() )
                  .log( Level.INFO, e.getLocalizedMessage(), e );
            setExportedStatus( EXPORT_ERROR, e.getLocalizedMessage() );
        }
    }

}
